@@ -10,6 +10,8 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
+import typing as ty
+
 from openstack.baremetal.v1 import _common
 from openstack.baremetal.v1 import allocation as _allocation
 from openstack.baremetal.v1 import chassis as _chassis
@@ -64,7 +66,7 @@ class Proxy(proxy.Proxy):
             error_message="No {resource_type} found for {value}".format(
                 resource_type=resource_type.__name__, value=value
             ),
-            **kwargs
+            **kwargs,
         )
 
     def chassis(self, details=False, **query):
@@ -118,7 +120,7 @@ class Proxy(proxy.Proxy):
 
         :param str name_or_id: The ID of a chassis.
         :param bool ignore_missing: When set to ``False``, an exception of
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the chassis does not exist.  When set to `True``, None will
             be returned when attempting to find a nonexistent chassis.
         :returns: One :class:`~openstack.baremetal.v1.chassis.Chassis` object
@@ -136,7 +138,7 @@ class Proxy(proxy.Proxy):
         :param fields: Limit the resource fields to fetch.
 
         :returns: One :class:`~openstack.baremetal.v1.chassis.Chassis`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound` when no
+        :raises: :class:`~openstack.exceptions.NotFoundException` when no
             chassis matching the name or ID could be found.
         """
         return self._get_with_fields(_chassis.Chassis, chassis, fields=fields)
@@ -172,7 +174,7 @@ class Proxy(proxy.Proxy):
         :param chassis: The value can be either the ID of a chassis or
             a :class:`~openstack.baremetal.v1.chassis.Chassis` instance.
         :param bool ignore_missing: When set to ``False``, an exception
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the chassis could not be found. When set to ``True``, no
             exception will be raised when attempting to delete a non-existent
             chassis.
@@ -206,7 +208,7 @@ class Proxy(proxy.Proxy):
             :class:`~openstack.baremetal.v1.driver.Driver` instance.
 
         :returns: One :class:`~openstack.baremetal.v1.driver.Driver`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound` when no
+        :raises: :class:`~openstack.exceptions.NotFoundException` when no
             driver matching the name could be found.
         """
         return self._get(_driver.Driver, driver)
@@ -218,7 +220,7 @@ class Proxy(proxy.Proxy):
             :class:`~openstack.baremetal.v1.driver.Driver` instance.
 
         :returns: One :dict: of vendor methods with corresponding usages
-        :raises: :class:`~openstack.exceptions.ResourceNotFound` when no
+        :raises: :class:`~openstack.exceptions.NotFoundException` when no
             driver matching the name could be found.
         """
         driver = self.get_driver(driver)
@@ -272,6 +274,7 @@ class Proxy(proxy.Proxy):
               ``provision_state``.
             * ``resource_class``: Only return those with the specified
               ``resource_class``.
+            * ``shard``: Only return nodes matching the supplied shard key.
             * ``sort_dir``: Sorts the response by the requested sort direction.
               A valid value is ``asc`` (ascending) or ``desc``
               (descending). Default is ``asc``. You can specify multiple
@@ -309,7 +312,7 @@ class Proxy(proxy.Proxy):
 
         :param str name_or_id: The name or ID of a node.
         :param bool ignore_missing: When set to ``False``, an exception of
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the node does not exist.  When set to `True``, None will
             be returned when attempting to find a nonexistent node.
         :returns: One :class:`~openstack.baremetal.v1.node.Node` object
@@ -327,7 +330,7 @@ class Proxy(proxy.Proxy):
         :param fields: Limit the resource fields to fetch.
 
         :returns: One :class:`~openstack.baremetal.v1.node.Node`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound` when no
+        :raises: :class:`~openstack.exceptions.NotFoundException` when no
             node matching the name or ID could be found.
         """
         return self._get_with_fields(_node.Node, node, fields=fields)
@@ -339,7 +342,7 @@ class Proxy(proxy.Proxy):
             :class:`~openstack.baremetal.v1.node.Node` instance.
 
         :returns: The node inventory
-        :raises: :class:`~openstack.exceptions.ResourceNotFound` when no
+        :raises: :class:`~openstack.exceptions.NotFoundException` when no
             inventory could be found.
         """
         res = self._get_resource(_node.Node, node)
@@ -455,7 +458,7 @@ class Proxy(proxy.Proxy):
         :return: The node boot device
         """
         res = self._get_resource(_node.Node, node)
-        return res.get_boot_device()
+        return res.get_boot_device(self)
 
     def set_node_boot_device(self, node, boot_device, persistent=False):
         """Set node boot device
@@ -478,7 +481,7 @@ class Proxy(proxy.Proxy):
         :return: The node boot device
         """
         res = self._get_resource(_node.Node, node)
-        return res.get_supported_boot_devices()
+        return res.get_supported_boot_devices(self)
 
     def set_node_boot_mode(self, node, target):
         """Make a request to change node's boot mode
@@ -701,7 +704,7 @@ class Proxy(proxy.Proxy):
         :param node: The value can be either the name or ID of a node or
             a :class:`~openstack.baremetal.v1.node.Node` instance.
         :param bool ignore_missing: When set to ``False``, an exception
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the node could not be found. When set to ``True``, no
             exception will be raised when attempting to delete a non-existent
             node.
@@ -772,7 +775,7 @@ class Proxy(proxy.Proxy):
 
         :param str name_or_id: The ID of a port.
         :param bool ignore_missing: When set to ``False``, an exception of
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the port does not exist.  When set to `True``, None will
             be returned when attempting to find a nonexistent port.
         :returns: One :class:`~openstack.baremetal.v1.port.Port` object
@@ -790,7 +793,7 @@ class Proxy(proxy.Proxy):
         :param fields: Limit the resource fields to fetch.
 
         :returns: One :class:`~openstack.baremetal.v1.port.Port`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound` when no
+        :raises: :class:`~openstack.exceptions.NotFoundException` when no
             port matching the name or ID could be found.
         """
         return self._get_with_fields(_port.Port, port, fields=fields)
@@ -826,7 +829,7 @@ class Proxy(proxy.Proxy):
         :param port: The value can be either the ID of a port or
             a :class:`~openstack.baremetal.v1.port.Port` instance.
         :param bool ignore_missing: When set to ``False``, an exception
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the port could not be found. When set to ``True``, no
             exception will be raised when attempting to delete a non-existent
             port.
@@ -891,7 +894,7 @@ class Proxy(proxy.Proxy):
 
         :param str name_or_id: The name or ID of a portgroup.
         :param bool ignore_missing: When set to ``False``, an exception of
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the port group does not exist.  When set to `True``, None will
             be returned when attempting to find a nonexistent port group.
         :returns: One :class:`~openstack.baremetal.v1.port_group.PortGroup`
@@ -909,7 +912,7 @@ class Proxy(proxy.Proxy):
         :param fields: Limit the resource fields to fetch.
 
         :returns: One :class:`~openstack.baremetal.v1.port_group.PortGroup`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound` when no
+        :raises: :class:`~openstack.exceptions.NotFoundException` when no
             port group matching the name or ID could be found.
         """
         return self._get_with_fields(
@@ -951,7 +954,7 @@ class Proxy(proxy.Proxy):
             :class:`~openstack.baremetal.v1.port_group.PortGroup`
             instance.
         :param bool ignore_missing: When set to ``False``, an exception
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the port group could not be found. When set to ``True``, no
             exception will be raised when attempting to delete a non-existent
             port group.
@@ -963,7 +966,15 @@ class Proxy(proxy.Proxy):
             _portgroup.PortGroup, port_group, ignore_missing=ignore_missing
         )
 
-    def attach_vif_to_node(self, node, vif_id, retry_on_conflict=True):
+    def attach_vif_to_node(
+        self,
+        node: ty.Union[_node.Node, str],
+        vif_id: str,
+        retry_on_conflict: bool = True,
+        *,
+        port_id: ty.Optional[str] = None,
+        port_group_id: ty.Optional[str] = None,
+    ) -> None:
         """Attach a VIF to the node.
 
         The exact form of the VIF ID depends on the network interface used by
@@ -973,17 +984,29 @@ class Proxy(proxy.Proxy):
 
         :param node: The value can be either the name or ID of a node or
             a :class:`~openstack.baremetal.v1.node.Node` instance.
-        :param string vif_id: Backend-specific VIF ID.
+        :param vif_id: Backend-specific VIF ID.
         :param retry_on_conflict: Whether to retry HTTP CONFLICT errors.
             This can happen when either the VIF is already used on a node or
             the node is locked. Since the latter happens more often, the
             default value is True.
-        :return: ``None``
+        :param port_id: The UUID of the port to attach the VIF to. Only one of
+            port_id or port_group_id can be provided.
+        :param port_group_id: The UUID of the portgroup to attach to. Only one
+            of port_group_id or port_id can be provided.
+        :return: None
         :raises: :exc:`~openstack.exceptions.NotSupported` if the server
             does not support the VIF API.
+        :raises: :exc:`~openstack.exceptions.InvalidRequest` if both port_id
+            and port_group_id are provided.
         """
         res = self._get_resource(_node.Node, node)
-        res.attach_vif(self, vif_id, retry_on_conflict=retry_on_conflict)
+        res.attach_vif(
+            self,
+            vif_id=vif_id,
+            retry_on_conflict=retry_on_conflict,
+            port_id=port_id,
+            port_group_id=port_group_id,
+        )
 
     def detach_vif_from_node(self, node, vif_id, ignore_missing=True):
         """Detach a VIF from the node.
@@ -996,7 +1019,7 @@ class Proxy(proxy.Proxy):
             a :class:`~openstack.baremetal.v1.node.Node` instance.
         :param string vif_id: Backend-specific VIF ID.
         :param bool ignore_missing: When set to ``False``
-            :class:`~openstack.exceptions.ResourceNotFound` will be
+            :class:`~openstack.exceptions.NotFoundException` will be
             raised when the VIF does not exist. Otherwise, ``False``
             is returned.
         :return: ``True`` if the VIF was detached, otherwise ``False``.
@@ -1074,7 +1097,7 @@ class Proxy(proxy.Proxy):
         :param fields: Limit the resource fields to fetch.
 
         :returns: One :class:`~openstack.baremetal.v1.allocation.Allocation`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound` when no
+        :raises: :class:`~openstack.exceptions.NotFoundException` when no
             allocation matching the name or ID could be found.
         """
         return self._get_with_fields(
@@ -1114,7 +1137,7 @@ class Proxy(proxy.Proxy):
         :param allocation: The value can be the name or ID of an allocation or
             a :class:`~openstack.baremetal.v1.allocation.Allocation` instance.
         :param bool ignore_missing: When set to ``False``, an exception
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the allocation could not be found. When set to ``True``, no
             exception will be raised when attempting to delete a non-existent
             allocation.
@@ -1166,7 +1189,7 @@ class Proxy(proxy.Proxy):
             :class:`~openstack.baremetal.v1.node.Node` instance.
         :param trait: trait to remove from the node.
         :param bool ignore_missing: When set to ``False``, an exception
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the trait could not be found. When set to ``True``, no
             exception will be raised when attempting to delete a non-existent
             trait.
@@ -1242,6 +1265,16 @@ class Proxy(proxy.Proxy):
         res = self._get_resource(_node.Node, node)
         return res.set_traits(self, traits)
 
+    def list_node_firmware(self, node):
+        """Lists firmware components for a node.
+
+        :param node: The value can be the name or ID of a node or a
+            :class:`~openstack.baremetal.v1.node.Node` instance.
+        :returns: A list of the node's firmware components.
+        """
+        res = self._get_resource(_node.Node, node)
+        return res.list_firmware(self)
+
     def volume_connectors(self, details=False, **query):
         """Retrieve a generator of volume_connector.
 
@@ -1299,7 +1332,7 @@ class Proxy(proxy.Proxy):
         :param str vc_id: The ID of a volume connector.
 
         :param bool ignore_missing: When set to ``False``, an exception of
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the volume connector does not exist.  When set to `True``,
             None will be returned when attempting to find a nonexistent
             volume connector.
@@ -1324,7 +1357,7 @@ class Proxy(proxy.Proxy):
 
         :returns: One
             :class: `~openstack.baremetal.v1.volume_connector.VolumeConnector`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound` when no
+        :raises: :class:`~openstack.exceptions.NotFoundException` when no
             volume_connector matching the name or ID could be found.`
         """
         return self._get_with_fields(
@@ -1374,7 +1407,7 @@ class Proxy(proxy.Proxy):
             :class:`~openstack.baremetal.v1.volume_connector.VolumeConnector`
             instance.
         :param bool ignore_missing: When set to ``False``, an exception
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the volume_connector could not be found.
             When set to ``True``, no exception will be raised when
             attempting to delete a non-existent volume_connector.
@@ -1446,7 +1479,7 @@ class Proxy(proxy.Proxy):
         :param str vt_id: The ID of a volume target.
 
         :param bool ignore_missing: When set to ``False``, an exception of
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the volume connector does not exist.  When set to `True``,
             None will be returned when attempting to find a nonexistent
             volume target.
@@ -1469,7 +1502,7 @@ class Proxy(proxy.Proxy):
 
         :returns: One
             :class:`~openstack.baremetal.v1.volume_target.VolumeTarget`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound` when no
+        :raises: :class:`~openstack.exceptions.NotFoundException` when no
             volume_target matching the name or ID could be found.`
         """
         return self._get_with_fields(
@@ -1516,7 +1549,7 @@ class Proxy(proxy.Proxy):
             :class:`~openstack.baremetal.v1.volume_target.VolumeTarget`
             instance.
         :param bool ignore_missing: When set to ``False``, an exception
-            :class:`~openstack.exceptions.ResourceNotFound` will be raised
+            :class:`~openstack.exceptions.NotFoundException` will be raised
             when the volume_target could not be found.
             When set to ``True``, no exception will be raised when
             attempting to delete a non-existent volume_target.
@@ -1584,7 +1617,7 @@ class Proxy(proxy.Proxy):
             instance.
 
         :param bool ignore_missing: When set to ``False``,
-            an exception:class:`~openstack.exceptions.ResourceNotFound`
+            an exception:class:`~openstack.exceptions.NotFoundException`
             will be raised when the deploy_template
             could not be found.
             When set to ``True``, no
@@ -1615,7 +1648,7 @@ class Proxy(proxy.Proxy):
 
         :returns: One
             :class:`~openstack.baremetal.v1.deploy_templates.DeployTemplate`
-        :raises: :class:`~openstack.exceptions.ResourceNotFound`
+        :raises: :class:`~openstack.exceptions.NotFoundException`
             when no deployment template matching the name or
             ID could be found.
         """
@@ -1662,7 +1695,7 @@ class Proxy(proxy.Proxy):
 
         :returns: One :class:`~openstack.baremetal.v1.conductor.Conductor`
 
-        :raises: :class:`~openstack.exceptions.ResourceNotFound` when no
+        :raises: :class:`~openstack.exceptions.NotFoundException` when no
             conductor matching the name could be found.
         """
         return self._get_with_fields(
